/**
 *
 * @typedef LoggerOptions
 * @param {array<number>} [onlyTypes] Типы логов для записи, по умолчанию все
 * @param {array<number>} [onlyLevels] Уровни логов для записи, по умолчанию все
 * @param {boolean} [isDebug] Если true, то логи уровня debug записываются
 * @param {string} [format] Формат логов для записи
 * @param {number} [maxParamsLength] Максимальная длина поля params
 */

require('sa-prototype');

const uuidv4 = require('uuid/v4');

const LOG_TYPE = {
  SYSTEM: 0,
  APPLICATION: 1,
  REQUEST: 3,
  ITERATION: 4,
  EVENT: 5,
  MODEL: 6,
  DB: 7,
  CACHE: 8,
  AMQP: 9
};

/**
 * Максимальная длина св-ва params в приведенного к строке, измеряется в кол-ве символов
 */
const MAX_PARAMS_LENGTH = 15000;

const LOG_LEVEL = {
  DUMP: 1,
  DEBUG: 2,
  INFO: 5,
  WARNING: 6,
  ERROR: 8,
  FATAL: 9,
};
/**
 * Форматы логов
 * json - json string
 * loghouse - prefix + json string + postfix
 */
const LOG_FORMAT = {
  JSON: 'json',
  LOGHOUSE: 'loghouse'
};

const APP_INFO_KEYS = ['ver', 'stage', 'port', 'uptime'];
const REQ_INFO_KEYS = ['rKey', 'rDp', 'rD', 'rDn', 'rU', 'rUp', 'rIP', 'rG', 'rCa', 'rCv'];

/**
 * Приводит аргумент params к json строке.
 * Если не получилось привести к json, то возвращаем сообщение об ошибке приведения.
 * @param {object|string} params
 * @returns {object} {string, boolean}
 */
const convertParamsToJSONString = (params) => {
  let res = '';
  let isFailed = true;

  try {
    isFailed = false;
    res = JSON.stringify(Object(params));
  } catch (e) {
    // пытаемся получить хоть что то
    const paramsByKeys = {};

    for (const key in params) {
      try {
        paramsByKeys[key] = JSON.stringify(params[key]);
      } catch (err) {
        paramsByKeys[key] = { msg: e.message, stack: e.stack };
      }
    }

    try {
      res = JSON.stringify(paramsByKeys);
    } catch (err) {
      res = JSON.stringify({ msg: 'Error formatting log parameters', stack: err.stack });
    }
  }

  return { res, isFailed };
};

/**
 * Пишем в stdout процесса лог в указанном формате
 * @param {string} logString
 * @param {string} format
 */
const writeLog = (logString, format, isError = false) => {
  let log;

  if (format === LOG_FORMAT.LOGHOUSE) {
    log = `###LogBegin###${logString}###LogEnd###\n`;
  } else if (format === LOG_FORMAT.JSON) {
    log = `###LogBegin###${logString}###LogEnd###\n`;
  } else {
    log = `###LogBegin###${logString}###LogEnd###\n`;
  }

  if (isError === false) {
    process.stdout.write(log);
  } else {
    process.stderr.write(log);
  }
};

const filter = (items, kind) => {
  items = items || Object.values(kind);
  const flags = {};
  for (const el of items) flags[el] = true;

  return flags;
};

module.exports = class Logger {
  static get TYPE() {
    return LOG_TYPE;
  }

  static get LVL() {
    return LOG_LEVEL;
  }

  static get FORMAT() {
    return LOG_FORMAT;
  }

  /**
   *
   * @param {*} env
   * @param {number} type
   * @param {LoggerOptions} [options]
   */
  constructor(env = {}, type = LOG_TYPE.APPLICATION, options = {}) {
    const { onlyTypes, onlyLevels, isDebug, format, maxParamsLength } = options;
    // for copy
    this.options = options;
    this.type = type;
    this.isDebug = isDebug || false;
    this.onlyTypes = filter(onlyTypes, LOG_TYPE);
    this.onlyLevels = filter(onlyLevels, LOG_LEVEL);
    this.format = format || LOG_FORMAT.LOGHOUSE;
    this.maxParamsLength = Number.isInteger(maxParamsLength) ? maxParamsLength : MAX_PARAMS_LENGTH;

    this.env = env;
    this.env.stage = env.stage || process.env.STAGE || 'dev';
    this.env.uptime = env.uptime || Date.now();
  }

  copy(env = {}, type) {
    return new Logger({ ...this.env, ...env }, type, this.options);
  }

  dump(message, params, isAlert = false) {
    this.write(message, params, LOG_LEVEL.DUMP, this.type, isAlert);
  }

  debug(message, params, isAlert = false) {
    this.write(message, params, LOG_LEVEL.DEBUG, this.type, isAlert);
  }

  info(message, params, isAlert = false) {
    this.write(message, params, LOG_LEVEL.INFO, this.type, isAlert);
  }

  warning(message, params, isAlert = false) {
    this.write(message, params, LOG_LEVEL.WARNING, this.type, isAlert);
  }

  error(message, params, isAlert = false) {
    if (typeof message === 'string') {
      this.write(message, params, LOG_LEVEL.ERROR, this.type, isAlert);
    } else {
      this.e(message, params, isAlert);
    }
  }

  fatal(message, params, isAlert = false) {
    this.write(message, params, LOG_LEVEL.FATAL, this.type, isAlert);
  }

  /**
   * @description Можно указать любой тип и уровень ошбки, окружение будет унаследовано от экземпляра логгера
   * @param message
   * @param params
   * @param lvl
   * @param type
   * @param isAlert
   */
  custom(message, params, lvl, type, isAlert = false) {
    this.write(message, params, Number(lvl), Number(type), isAlert);
  }

  /**
   * @description Метод для ленивых можно просто бросить ошибку
   * @param {string|Error} err
   * @param {*} [params]
   * @param isAlert
   */
  e(err, params, isAlert = false) {
    if (params && typeof params === 'object' && !Array.isArray(params)) {
      params.stack = err.stack;
    } else {
      params = {
        stack: err.stack,
        pars: params
      };
    }

    if (err.params) {
      params.mergeFrom(err.params);
    }

    params.status = err.status || 0;

    this.write(err.message, params, LOG_LEVEL.ERROR, this.type, isAlert);
  }

  write(message, params, lvl = LOG_LEVEL.DEBUG, type, isAlert = false) {
    if (lvl === LOG_LEVEL.DEBUG && !this.isDebug) {
      return;
    }

    if (!this.onlyTypes[type] || !this.onlyLevels[lvl]) {
      return;
    }

    try {
      const now = Date.now();
      const logObject = { ...this.env, type, msg: message, lvl, ts: now };

      logObject.rKey = logObject.rKey || `${now / 1000}-${uuidv4()}`;

      if (params) {
        const { res: pars, isFailed } = convertParamsToJSONString(params);
        logObject.pars = (this.maxParamsLength !== 0) ? pars.substring(0, this.maxParamsLength) : pars;

        // если не смогли правильно преобразовать параметры,
        // то принудительно оповещаем об этом
        if (isFailed) {
          isAlert = true;
        }
      }

      if (isAlert) {
        logObject.alert = 1;
      }

      const { res: logString } = convertParamsToJSONString(logObject);

      writeLog(logString, this.format, lvl >= LOG_LEVEL.WARNING);
    } catch (e) {
      console.log(e);
    }
  }

  static get appInfoKeys() {
    return APP_INFO_KEYS;
  }

  static get reqInfoKeys() {
    return REQ_INFO_KEYS;
  }

  get appInfo() {
    return this.env.pickKeys(Logger.appInfoKeys);
  }

  get reqInfo() {
    return this.env.pickKeys(Logger.reqInfoKeys);
  }

  get rKey() {
    return this.env.rKey;
  }

  setEnv(key, val) {
    this.env[key] = val;
  }

  getEnv(key) {
    return this.env[key];
  }

  updateReqCtxByEnv(ctx) {
    for (const key of Logger.reqInfoKeys) {
      const val = this.env[key];

      if (val || val === 0) {
        ctx.set(key, val);
      }
    }
  }

  updateByDevice(device = {}) {
    this.env.rDp = device.devPlatform;
    this.env.rD = device.ID;
    this.env.rDn = device.devName;
    this.env.rU = device.userID;

    this.env.rCa = device.appName;
    this.env.rCv = device.appVer;
  }

  updateByUser(user = {}) {
    this.env.rU = user.ID;
    this.env.rUp = user.phone;
  }
};
