/**
 * Created by Nikolay Ilin (na.iline@me.com) on 15.12.2017.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

const _ = require('lodash');
const uuidv4 = require('uuid/v4');

const Logger = require('./Logger');
const deps = require('./Dependencies');
const ErrorWrp = require('./Error');
const { asyncSleep } = require('./helpers');

const AppError = new ErrorWrp('QWorker');

module.exports = (dirname) => {
  const pkg = require(`${dirname}/app.json`);

  let logger = null;

  const iterErrors = {};
  const config = {
    iterationSleep: 30, // seconds
    selfCheckInterval: 60 // seconds
  };

  let state;
  let workerLoop = true;
  let runsCount = 1;

  async function selfCheckingLoop() {
    const runID = uuidv4();

    let selfCheckDate = Date.now() + (config.selfCheckInterval * 1000);

    while (workerLoop) {
      state.logger.setEnv('rKey', `${runID}#${runsCount}`);
      state.logger.debug(`Start #${runsCount}`);
      state.logger.setEnv('rts', Date.now());

      if (selfCheckDate < Date.now()) {
        try {
          await selfCheck();
        } catch (err) {
          logger.error(err);
          await asyncSleep(15);
          await selfCheck();
        }

        selfCheckDate = Date.now() + (config.selfCheckInterval * 1000);
      }

      state.logger.debug(`End #${runsCount}`, { sleep: config.iterationSleep, nextCheck: (new Date(selfCheckDate).toDtString()) });

      await asyncSleep(config.iterationSleep);

      runsCount++;
    }

    await end();
  }

  async function start(appConfig, appState = {}) {
    try {
      _.merge(config, appConfig);

      logger = new Logger(
        {
          app: pkg.name,
          ver: pkg.version,
          stage: config.stage,
        },
        Logger.TYPE.APPLICATION,
        config.loggerOptions || {}
      );

      state = appState;
      state.config = config;

      logger.debug('starting...');

      const selfCheckResult = await deps.connect(state, config, logger);

      await selfCheck(selfCheckResult);

      // GRACEFUL SHUTDOWN
      process.on('SIGINT', shutdownHandler);

      // GRACEFUL SHUTDOWN
      process.on('SIGTERM', shutdownHandler);

      logger.info('⬆️ APP Started', null, true);

      state.logger = logger.copy({}, Logger.TYPE.EVENT);
      state.iterError = AppError;

      selfCheckingLoop(); // if to try use "await" we will get lock

      return appState;
    } catch (e) {
      await end(e);

      return null;
    }
  }

  async function end(err) {
    workerLoop = false;

    await deps.end(state);
    // todo логгер может не создаться на старте
    logger.info('⬇️App stopped!', null, true);
    err && logger.error(err);

    process.exit(err ? 1 : 0);
  }

  async function selfCheck(selfCheckResult) {
    const checkState = selfCheckResult || await deps.check(state, iterErrors) || {};

    checkState.ts = Date.now();
    checkState.sleep = config.iterationSleep;
    checkState.iterCount = runsCount;

    await state.cache.set(`${logger.getEnv('app')}-state`, checkState);
  }

  //--------------------------------------------------------

  function shutdownHandler() {
    workerLoop = false;

    // Force close
    setTimeout(async () => {
      logger.error('Forcing server close !!!');
      await end();
    }, 5000);
  }

  return { start, end, asyncSleep };
};
