// todo требовать логгер при подключении или пробрасывать конфиг для нового логгера
/**
 * Created by Nikolay Ilin (na.iline@me.com) on 23/01/2019.
 * Copyrights by Soft-Artel LIMITED
 * www.soft-artel.com
 */

require('sa-prototype');

const RedisCli = require('ioredis');

const Logger = require('./Logger');
const ErrorWrp = require('./Error');

const Error = new ErrorWrp('Redis');

class Redis {
  static async connect(config, parentLogger = {}) {
    const redis = new Redis(config, parentLogger);
    await redis.connect();

    return redis;
  }

  // ------------------------------------
  // LIFE CIRCLE

  constructor(config, parentLogger) {
    this.logger = parentLogger.copy({ uptime: Date.now() }, Logger.TYPE.CACHE) || new Logger({ uptime: Date.now() }, Logger.TYPE.CACHE);

    // https://github.com/luin/ioredis/blob/HEAD/API.md#new_Redis
    this.config = {
      port: 6379,
      host: 'localhost',
      db: 0,
      maxRetriesPerRequest: null // Every command will wait forever until the connection is alive again
    };

    this.config.mergeFrom(config);

    this.requestsCount = 0;
  }

  // ------------------------------------

  async connect() {
    try {
      this.redis = new RedisCli(this.config);
      const state = await new Promise(resolve => this.redis.on('ready', () => resolve('is OK!')));
      this.logger.debug(`Redis connection ${state}`);
    } catch (err) {
      Error.throw(err, 500, this.config);
    }
  }

  // ------------------------------------

  async end() {
    try {
      await this.redis.disconnect();
      this.logger.debug(`Redis connection closed, ${this.requestsCount} requests complete`);
    } catch (err) {
      Error.throw(err, 500, this.config);
    }
  }

  // -----------------
  // COMMAND WRAPPER

  async cmd(cmd = 'get', key, data = null, ttl = null) {
    this.requestsCount++;

    // this.logger.dump(`Redis ${cmd} #${this.requestsCount} key:${key}`, data);

    if (cmd === 'set' && ttl) {
      cmd = 'setex';
    }

    try {
      // let cmdTime = Date.now();
      let res;

      if (data === null) {
        res = await this.redis[cmd](key);

        try {
          res = JSON.parse(res);
        } catch (e) {
          // maybe some code here
        }
      }

      if (data !== null && ttl === null) {
        res = await this.redis[cmd](key, data);
      }

      if (data !== null && ttl !== null) {
        res = await this.redis[cmd](key, ttl, data);
      }

      // cmdTime = Date.now() - cmdTime;

      // this.logger.dump(`Redis #${this.requestsCount} ${cmd}${ttl ? ` ttl:${ttl}` : ''} complete by ${String.humanizeMsDuration(cmdTime)}`);

      return res;
    } catch (err) {
      Error.throw(err, 500, { cmd, key });
    }

    return null;
  }

  // ------------------------------------
  // COMMANDS!

  async keys(mask = '*') {
    return await this.cmd('keys', mask);
  }

  // -----------------

  async set(key, data, ttl = null) {
    if (data && typeof data !== 'string') {
      data = JSON.stringify(data);
    }

    return await this.cmd('set', key, data, ttl);
  }

  // -----------------

  async setex(key, data, ttl = 5) {
    return await this.set(key, data, ttl);
  }

  // -----------------

  async get(key) {
    return await this.cmd('get', key);
  }

  // -----------------

  async del(keys) {
    if (Array.isArray(keys)) {
      keys = keys.join(' ');
    }

    return await this.cmd('del', keys);
  }

  // -----------------

  async llen(key) {
    return await this.cmd('llen', key);
  }

  // -----------------

  async push(key, items, type = 'r') {
    // Insert all the specified values at the
    // type = 'l' - START of the list stored at key.
    // type = 'r' - END of the list stored at key.

    const cmd = `${type}push`;

    if (!Array.isArray(items)) {
      items = [items];
    }

    const data = [];

    items.forEach((item) => {
      if (typeof item === 'string') {
        data.push(item);
      } else {
        data.push(JSON.stringify(item));
      }
    });

    return await this.cmd(cmd, key, data);
  }

  // -----------------

  async pop(key, type = 'l') {
    // Removes and returns the
    // type = 'l' FIRST element of the list stored at key.
    // type = 'r' LAST  element of the list stored at key.

    return await this.cmd(`${type}pop`, key);
  }

  // -----------------

  async lindex(key, index = 0) {
    let res = await this.cmd('lindex', key, index);

    try {
      res = JSON.parse(res);
    } catch (e) {
      // maybe some code here
    }

    return res;
  }

  // -----------------

  async lrange(key, from = 0, to = 1000) {
    let res = await this.cmd('lrange', key, to, from);

    if (typeof res === 'string') {
      res = [res];
    }

    const items = [];

    res.forEach((item) => {
      try {
        items.push(JSON.parse(item));
      } catch (e) {
        items.push(item);
      }
    });

    return items;
  }
}

module.exports = Redis;
